/*
###############################################################################
#                                                                             #
#   gfslogger v0.9                                                            #
#   Subscribes to Mac OS X 10.4 fsevents and displays all filesystem changes. #
#                                                                             #
#   Copyright (C) 2005 Rian Hunter [rian at thelaststop point net]            #
#                                                                             #
#   This program is free software; you can redistribute it and/or modify      #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#                                                                             #
#   This program is distributed in the hope that it will be useful,           #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of            #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
#   GNU General Public License for more details.                              #
#                                                                             #
#   You should have received a copy of the GNU General Public License         #
#   along with this program; if not, write to the Free Software	              #
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA #
#                                                                             #
###############################################################################
 */

/*
 * ========================================
 * Two changes for my purposes:
 * 1/ Only outputs file paths.
 * 2/ Stamps events with an approximate time of generation.
 * 3/ Automatically kills itself if stdout dies.
 *
 * RbrtPntn - 20 August 2005
 * ========================================
 */

// for open(2)
#include <fcntl.h>

// for ioctl(2)
#include <sys/ioctl.h>
#include <sys/sysctl.h>

// for read(2)
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

// for printf(3)
#include <stdio.h>

// for exit(3)
#include <stdlib.h>

// for strncpy(3)
#include <string.h>

// for getpwuid(3)
#include <pwd.h>

// for getgrgid(3)
#include <grp.h>

// for S_IS*(3)
#include <sys/stat.h>

#include <time.h> 

// duh.
#include "fsevents.h"

//static void die(int);
static void process_event_data(void *, int);
//static void get_process_name(pid_t, char *, int);
//static void get_mode_string(int32_t, char *);
//static char *get_vnode_type(int32_t);


char large_buf[0x2000];

// activates self as fsevent listener and displays fsevents
// must be run as root!! (at least on Mac OS X 10.4)
int main() {
	int newfd, fd, n;
	signed char event_list[FSE_MAX_EVENTS];
	fsevent_clone_args retrieve_ioctl;
	
	event_list[FSE_CREATE_FILE]         = FSE_REPORT;
	event_list[FSE_DELETE]              = FSE_REPORT;
	event_list[FSE_STAT_CHANGED]        = FSE_REPORT;
	event_list[FSE_RENAME]              = FSE_REPORT;
	event_list[FSE_CONTENT_MODIFIED]    = FSE_REPORT;
	event_list[FSE_EXCHANGE]            = FSE_REPORT;
	event_list[FSE_FINDER_INFO_CHANGED] = FSE_REPORT;
	event_list[FSE_CREATE_DIR]          = FSE_REPORT;
	event_list[FSE_CHOWN]               = FSE_REPORT;
	
	fd = open("/dev/fsevents", 0, 2);
	if (fd < 0) {
		exit(1);
	}
	
	retrieve_ioctl.event_list = event_list;
	retrieve_ioctl.num_events = sizeof(event_list);
	retrieve_ioctl.event_queue_depth = 0x400;
	retrieve_ioctl.fd = &newfd;
	
	if (ioctl(fd, FSEVENTS_CLONE, &retrieve_ioctl) < 0) {
		exit(1);
	}
	close(fd);
	
	fprintf(stderr, "Tool - gfslogger running...\n");
	
	// note: you must read at least 2048 bytes at a time on this fd, to get data.
	// also you read quickly! newer events can be lost in the internal kernel event
	// buffer if you take too long to get events. thats why buffer is so large:
	// fewer read calls.
	while ((n = read(newfd, large_buf, sizeof(large_buf))) > 0) {
		process_event_data(large_buf, n);
	}
	
	return 0;
}

void printPath(int32_t type, pid_t pid, char *path) {
	/* if we can't write to stdout anymore then kill ourself - either our parent died or the fd closed*/
	struct stat st;
  	if(fstat(1, &st) != 0) exit(0); /* exit nicely */
	
	/* only write file paths */
	if(!path[0] || (path[0] != '/')) return;
		
	//fprintf(stdout, "/%d:%ld:%ld%s\n", (int)type, (long)pid, when, path);
	fprintf(stdout, "%s\n", path);
	fflush(stdout);
}

/* event structure in mem:

event type: 4 bytes
event pid: sizeof(pid_t) (4 on darwin) bytes
args:
argtype: 2 bytes
arglen: 2 bytes
argdata: arglen bytes
lastarg:
argtype: 2 bytes = 0xb33f

*/

// parses the incoming event data and displays it in a friendly way
static void process_event_data(void *in_buf, int size) {
	int pos = 0;
	pid_t pid;
	u_int16_t argtype;
	u_int16_t arglen;
	
	//printf("=> received %d bytes\n", size);
	
	do {
		//printf("# Event\n");
		int32_t type = *((int32_t *) (in_buf + pos));
		//printf("  type           = ");
		switch (type) {
			case FSE_CREATE_FILE:
				//printf("CREATE FILE");
				break;
			case FSE_DELETE:
				//printf("DELETE");
				break;
			case FSE_STAT_CHANGED:
				//printf("STAT CHANGED");
				break;
			case FSE_RENAME:
				//printf("RENAME");
				break;
			case FSE_CONTENT_MODIFIED:
				//printf("CONTENT MODIFIED");
				break;
			case FSE_EXCHANGE:
				//printf("EXCHANGE");
				break;
			case FSE_FINDER_INFO_CHANGED:
				//printf("FINDER INFO CHANGED");
				break;
			case FSE_CREATE_DIR:
				//printf("CREATE DIR");
				break;
			case FSE_CHOWN:
				//printf("CHOWN");
				break;
			case FSE_INVALID: default:
				//printf("INVALID");
				return; // <----------we return if invalid type (give up on this data)
				break;
		}

		pos += 4;
		pid = *((pid_t *) (in_buf + pos));
		pos += sizeof(pid_t);
				
		while(1) {
			argtype = *((u_int16_t *) (in_buf + pos));
			pos += 2;
			
			if (FSE_ARG_DONE == argtype) {
				break;
			}
			
			arglen = *((u_int16_t *) (in_buf + pos));
			pos += 2;
			
			switch(argtype) {
				case FSE_ARG_VNODE:
					printPath(type, pid, in_buf + pos);
					break;
				case FSE_ARG_STRING:
					printPath(type, pid, in_buf + pos);
					break;
				case FSE_ARG_PATH: // not in kernel
					printPath(type, pid, in_buf + pos);
					break;
				case FSE_ARG_INT32:
					break;
				case FSE_ARG_INT64: // not supported in kernel yet
					break;
				case FSE_ARG_RAW: // just raw bytes, can't display
					break;
				case FSE_ARG_INO:
					break;
				case FSE_ARG_UID:
					break; 
				case FSE_ARG_DEV:
					break;
				case FSE_ARG_MODE:
					break;
				case FSE_ARG_GID:
					break; 
				default:
					return; // <----------we return if invalid type (give up on this data)
					break;
			}
			pos += arglen;
		}
		
	} while (pos < size);
	
	return;
}
